/**********************************************************************
 Freeciv - Copyright (C) 2005 - The Freeciv Project
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#include <stdarg.h>

#include "lua.h"
#include "lualib.h"
#include "tolua.h"

#include "hash.h"

#include "api_gen.h"

#include "api_types.h"
#include "api_store.h"
#include "script.h"

/* get 'struct signal_list' and related functions: */
#define SPECLIST_TAG signal
#define SPECLIST_TYPE char
#include "speclist.h"

#define signal_list_iterate(signallist, psignal) \
    TYPED_LIST_ITERATE(char, signallist, psignal)
#define signal_list_iterate_end  LIST_ITERATE_END

static struct {
  lua_State *state;

  struct hash_table *signals;
  struct hash_table *callbacks;
} script_data;

/* callback functions. */
static bool callback_status_get(const char *callback_name)
{
  struct hash_table *hash = script_data.callbacks;

  return (bool) hash_lookup_data(hash, callback_name);
}

static bool callback_status_set(const char *callback_name, bool enabled)
{
  struct hash_table *hash = script_data.callbacks;
  const void *key;

  if (hash_lookup(hash, callback_name, &key, NULL)) {
    hash_replace(hash, key, (void *) enabled);
    return FALSE;
  } else {
    hash_insert(hash, callback_name, (void *) enabled);
    return TRUE;
  }
}

static bool callback_invoke_valist(const char *callback_name, va_list args)
{
  lua_State *state = script_data.state;

  if (callback_status_get(callback_name)) {
    int nargs, nres;
    bool res = FALSE;

    /* the function name */
    lua_getglobal(state, callback_name);

    for (nargs = 0; ; nargs++) {
      int type;

      if ((type = va_arg(args, int)) == -1) {
	break;
      }

      switch (type) {
	case API_TYPE_INT:
	  {
	    int arg;

	    arg = va_arg(args, int);
	    tolua_pushnumber(state, (lua_Number)arg);
	  }
	  break;
	case API_TYPE_BOOL:
	  {
	    int arg;

	    arg = va_arg(args, int);
	    tolua_pushboolean(state, (bool)arg);
	  }
	  break;
	case API_TYPE_STRING:
	  {
	    const char *arg;

	    arg = va_arg(args, const char*);
	    tolua_pushstring(state, arg);
	  }
	  break;
	case API_TYPE_PLAYER:
	  {
	    Player *arg;

	    arg = va_arg(args, Player*);
	    tolua_pushusertype(state, (void*)arg, "Player");
	  }
	  break;
	case API_TYPE_CITY:
	  {
	    City *arg;

	    arg = va_arg(args, City*);
	    tolua_pushusertype(state, (void*)arg, "City");
	  }
	  break;
	case API_TYPE_UNIT:
	  {
	    Unit *arg;

	    arg = va_arg(args, Unit*);
	    tolua_pushusertype(state, (void*)arg, "Unit");
	  }
	  break;
	case API_TYPE_TILE:
	  {
	    Tile *arg;

	    arg = va_arg(args, Tile*);
	    tolua_pushusertype(state, (void*)arg, "Tile");
	  }
	  break;
      }
    }

    /* call the function with nargs arguments, return 1 results */
    if (lua_pcall(state, nargs, 1, 0) != 0) {
      return FALSE;
    }

    nres = lua_gettop(state);

    if (lua_isboolean(state, -1)) {
      res = lua_toboolean(state, -1);
      lua_pop(state, nres);

      if (res) {
	/* callback wants to stop running */
	callback_status_set(callback_name, FALSE);
	return TRUE;
      }
    } else {
      /* invalid type returned by callback */
      lua_pop(state, nres);
    }
  }
  return FALSE;
}

/* signal functions. */
static void script_signal_invoke_valist(const char *signal_name, va_list args)
{
  struct hash_table *hash;
  struct signal_list *signals;

  hash = script_data.signals;
  signals = hash_lookup_data(hash, signal_name);

  if (signals) {
    signal_list_iterate(signals, callback_name) {
      if (callback_invoke_valist(callback_name, args)) {
	signal_list_unlink(signals, callback_name);
      }
    } signal_list_iterate_end;
  }
}

void script_signal_emit_valist(const char *signal_name, va_list args)
{
  script_signal_invoke_valist(signal_name, args);
}

void script_signal_emit(const char *signal_name, ...)
{
  va_list args;

  va_start(args, signal_name);
  script_signal_invoke_valist(signal_name, args);
  va_end(args);
}

void script_signal_new_valist(const char *signal_name, va_list args)
{
  struct hash_table *hash;
  struct signal_list *list;
  char *name;

  hash = script_data.signals;
  list = signal_list_new();
  name = strdup(signal_name);

  if (!hash_insert(hash, name, list)) {
    signal_list_free(list);
    free(name);
  }
}

void script_signal_new(const char *signal_name, ...)
{
  va_list args;

  va_start(args, signal_name);
  script_signal_new_valist(signal_name, args);
  va_end(args);
}

void script_signal_connect(const char *signal_name, const char *callback_name)
{
  if (!signal_name) {
    script_error("nil string argument 'signal_name'.");
  } else if (!callback_name) {
    script_error("nil string argument 'callback_name'.");
  } else {
    struct signal_list *list;
    char *name;

    list = hash_lookup_data(script_data.signals, signal_name);
    name = strdup(callback_name);

    signal_list_append(list, name);
    if (!callback_status_set(name, TRUE)) {
      free(name);
    }
  }
}

/* internal api error function. */
void script_error(const char *message)
{
  luaL_error(script_data.state, message);
}

/* script functions. */
bool script_init(void)
{
  lua_State *state;

  state = lua_open();
  if (!state) {
    return FALSE;
  }

  script_data.state = state;

  luaopen_base(state);
  luaopen_string(state);
  luaopen_io(state);
  luaopen_debug(state);
  luaopen_table(state);

  tolua_api_open(state);

  api_store_init();

  script_data.signals = hash_new(hash_fval_string, hash_fcmp_string);
  script_data.callbacks = hash_new(hash_fval_string, hash_fcmp_string);

  return TRUE;
}

void script_free(void)
{
  hash_free(script_data.signals);
  hash_free(script_data.callbacks);

  api_store_free();
}

bool script_do_file(const char *filename)
{
  return lua_dofile(script_data.state, filename);
}

